
import scipy as sp


def fit(observed, modeled, initial_data, initial_params, evaluate, *vary,
        tol=0.01, forwards=None, backwards=None):
    '''
    Fit a model to some data.

    The model is parameterised by a namedtuple, with initial values initial_params.
    The evaluate() function takes some data and model parameters and generates the model (in the same data frame).
    The observed and modeled columns in the dataframe are then compared.

    forwards and backwards allow the model parameters to be transformed as necessary.
    forwards goes from the model to the constrained parameters; backwards reverses the transform.
    '''

    if forwards is None: forwards = lambda x: x
    if backwards is None: backwards = lambda x: x

    def objective(args):
        params = initial_params._replace(**backwards(dict(zip(vary, args))))
        data = evaluate(initial_data, params)
        # return sp.stats.chisquare(data[observed], data[modeled]).statistic
        delta = (data[observed] - data[modeled]).dropna()
        return sum(delta * delta)

    result = sp.optimize.minimize(objective,
                                  [forwards(initial_params._asdict())[name] for name in vary],
                                  method='Nelder-Mead', tol=tol)
    return initial_params._replace(**backwards(dict(zip(vary, result.x))))





from collections import Counter
from itertools import chain
from json import loads
from logging import getLogger

from scipy.interpolate import UnivariateSpline

from . import MultiProcCalculator, ActivityJournalCalculatorMixin, WaypointCalculatorMixin
from ..names import *
from ..waypoint import Chunks
from ...data.climb import find_climbs, Climb
from ...squeal import Constant, StatisticJournalFloat
from ...stoats.calculate.heart_rate import hr_zones_from_database

log = getLogger(__name__)
HR_MINUTES = (5, 10, 15, 20, 30, 60, 90, 120, 180)


def round_km():
    yield from range(5, 21, 5)
    yield from range(25, 76, 25)
    yield from range(100, 251, 50)
    yield from range(300, 1001, 100)


class TimeForDistance(Chunks):

    def __init__(self, waypoints, distance):
        super().__init__(waypoints)
        self.__distance = distance

    def times(self):
        for chunks in self.chunks():
            distance = sum(chunk.distance() for chunk in chunks)
            if distance > self.__distance:
                while chunks and distance - chunks[0].distance_delta() > self.__distance:
                    distance -= chunks[0].distance_delta()
                    self.drop_first(chunks)
                time = sum(chunk.time() for chunk in chunks)
                yield time * self.__distance / distance


class MedianHRForTime(Chunks):

    def __init__(self, waypoints, time, max_gap=None):
        super().__init__(waypoints)
        self.__time = time
        self.__max_gap = 0.01 * time if max_gap is None else max_gap
        log.debug('Will reject gaps > %ds' % int(self.__max_gap))

    def _max_gap(self, chunks):
        return max(c1[0].timespan.start - c2[0].timespan.finish
                   for c1, c2 in zip(list(chunks)[1:], chunks)).total_seconds()

    def heart_rates(self):
        for chunks in self.chunks():
            while len(chunks) > 1 and self._max_gap(chunks) > self.__max_gap:
                log.debug('Rejecting chunk because of gap (%ds)' % int(self._max_gap(chunks)))
                chunks.popleft()
            time = sum(chunk.time() for chunk in chunks)
            if time > self.__time:
                while chunks and time - chunks[0].time_delta() > self.__time:
                    time -= chunks[0].time_delta()
                    self.drop_first(chunks)
                heart_rates = list(sorted(chain(*(chunk.heart_rates() for chunk in chunks))))
                if heart_rates:
                    median = len(heart_rates) // 2
                    yield heart_rates[median]


class Totals(Chunks):

    def __init__(self, waypoints):
        super().__init__(waypoints)
        all_chunks = self.complete()
        self.distance = sum(chunk.distance() for chunk in all_chunks)
        self.time = sum(chunk.time() for chunk in all_chunks)


class Zones(Chunks):

    def __init__(self, waypoints, zones):
        super().__init__(waypoints)
        # this assumes record data are evenly distributed
        self.zones = []
        all_chunks = self.complete()
        counts = Counter()
        lower_limit = 0
        for zone, upper_limit in enumerate(zones):
            for chunk in all_chunks:
                for heart_rate in chunk.heart_rates():
                    if heart_rate is not None:
                        if lower_limit <= heart_rate < upper_limit:
                            counts[zone] += 1  # zero-based (incremented below)
            lower_limit = upper_limit
        total = sum(counts.values())
        if total:
            for zone in range(len(zones)):
                self.zones.append((zone + 1, counts[zone] / total))


class ActivityCalculator(ActivityJournalCalculatorMixin, WaypointCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, cost_calc=10, cost_write=1, climb=None, **kargs):
        self.climb_ref = climb
        super().__init__(*args, cost_calc=cost_calc, cost_write=cost_write, **kargs)

    def _names(self):
        return {HEART_RATE: 'heart_rate',
                DISTANCE: 'distance',
                RAW_ELEVATION: 'raw_elevation',
                ELEVATION: 'elevation'}

    def _get_loader(self, s):
        # no serial because we timetravel below
        return super()._get_loader(s, add_serial=False)

    def _calculate_results(self, s, ajournal, waypoints, loader):
        waypoints = self._fix_elevation(s, ajournal, waypoints, loader)
        totals = self._add_totals(s, ajournal, waypoints, loader)
        self._add_times_for_distance(s, ajournal, waypoints, loader)
        self._add_hr_stats(s, totals, ajournal, waypoints, loader)
        self._add_climbs(s, ajournal, waypoints, loader)

    def _add_totals(self, s, ajournal, waypoints, loader):
        totals = Totals(waypoints)
        loader.add(ACTIVE_DISTANCE, M, summaries(MAX, CNT, SUM, MSR), ajournal.activity_group, ajournal,
                   totals.distance, ajournal.start, StatisticJournalFloat)
        loader.add(ACTIVE_TIME, S, summaries(MAX, SUM, MSR), ajournal.activity_group, ajournal,
                   totals.time, ajournal.start, StatisticJournalFloat)
        loader.add(ACTIVE_SPEED, KMH, summaries(MAX, AVG, MSR), ajournal.activity_group, ajournal,
                   totals.distance * 3.6 / totals.time, ajournal.start, StatisticJournalFloat)
        return totals

    def _add_times_for_distance(self, s, ajournal, waypoints, loader):
        for target in round_km():
            times = list(sorted(TimeForDistance(waypoints, target * 1000).times()))
            if not times:
                break
            median = len(times) // 2
            loader.add(MEDIAN_KM_TIME % target, S, summaries(MIN, MSR), ajournal.activity_group, ajournal,
                       times[median], ajournal.start, StatisticJournalFloat)

    def _add_hr_stats(self, s, totals, ajournal, waypoints, loader):
        zones = hr_zones_from_database(s, ajournal.activity_group, ajournal.start)
        if zones:
            for (zone, frac) in Zones(waypoints, zones).zones:
                loader.add(PERCENT_IN_Z % zone, PC, None, ajournal.activity_group, ajournal,
                           100 * frac, ajournal.start, StatisticJournalFloat)
                loader.add(TIME_IN_Z % zone, S, None, ajournal.activity_group, ajournal,
                           frac * totals.time, ajournal.start, StatisticJournalFloat)
            for target in HR_MINUTES:
                heart_rates = sorted(MedianHRForTime(waypoints, target * 60).heart_rates(), reverse=True)
                if heart_rates:
                    loader.add(MAX_MED_HR_M % target, BPM, summaries(MAX, MSR), ajournal.activity_group, ajournal,
                           heart_rates[0], ajournal.start, StatisticJournalFloat)
        else:
            log.warning('No HR zones defined for %s or before' % ajournal.start)

    def _fix_elevation(self, s, ajournal, waypoints, loader):
        with_elevations = [waypoint for waypoint in waypoints if waypoint.raw_elevation != None]
        if len(with_elevations) > 4:
            fixed = []
            x = [waypoint.distance for waypoint in with_elevations]
            y = [waypoint.raw_elevation for waypoint in with_elevations]
            i = 1
            while i < len(x):
                if x[i-1] >= x[i]:
                    del x[i-1], y[i-1]
                else:
                    i += 1
            # the 7 here is from eyeballing various plots compared to other values
            # it seems better to smooth along the route rather that smooth the terrain model since
            # 1 - we expect the route to be smoother than the terrain in general (roads / tracks)
            # 2 - smoothing the 2d terrain is difficult to control and can give spikes
            # 3 - we better handle errors from mismatches between terrain model and position
            #     (think hairpin bends going up a mountainside)
            # the main drawbacks are
            # 1 - speed on loading
            # 2 - no guarantee of consistency between routes (or even on the same routine retracing a path)
            spline = UnivariateSpline(x, y, s=len(with_elevations) * 7)
            for waypoint in with_elevations:
                elevation = spline(waypoint.distance)
                loader.add(ELEVATION, M, None, ajournal.activity_group, ajournal,
                           elevation, waypoint.time, StatisticJournalFloat)
                fixed.append(waypoint._replace(elevation=elevation))
            return fixed
        else:
            return waypoints

    def _add_climbs(self, s, ajournal, waypoints, loader):
        climb = Climb(**loads(Constant.get(s, self.climb_ref).at(s).value))
        total_elevation = 0
        for lo, hi in find_climbs(waypoints, params=climb):
            up = hi.elevation - lo.elevation
            along = hi.distance - lo.distance
            time = (hi.time - lo.time).total_seconds()
            loader.add(CLIMB_ELEVATION, M, summaries(MAX, SUM, MSR), ajournal.activity_group, ajournal,
                       up, hi.time, StatisticJournalFloat)
            loader.add(CLIMB_DISTANCE, M, summaries(MAX, SUM, MSR), ajournal.activity_group, ajournal,
                       along, hi.time, StatisticJournalFloat)
            loader.add(CLIMB_TIME, S, summaries(MAX, SUM, MSR), ajournal.activity_group, ajournal,
                       time, hi.time, StatisticJournalFloat)
            loader.add(CLIMB_GRADIENT, PC, summaries(MAX, SUM, MSR), ajournal.activity_group, ajournal,
                       100 * up / along, hi.time, StatisticJournalFloat)
            total_elevation += up
        if total_elevation:
            loader.add(TOTAL_CLIMB, M, summaries(MAX, SUM, MSR), ajournal.activity_group, ajournal,
                       total_elevation, ajournal.start, StatisticJournalFloat)





import datetime as dt
from collections import namedtuple
from json import loads
from logging import getLogger
from math import exp

from sqlalchemy import desc, select, and_

from ch2.squeal import CompositeComponent
from . import UniProcCalculator, CompositeCalculatorMixin
from .heart_rate import HRImpulse
from ..names import MAX
from ...data.frame import _tables
from ...lib.date import local_date_to_time, time_to_local_date
from ...squeal import Constant, StatisticJournal, StatisticName, StatisticJournalFloat, ActivityJournal

log = getLogger(__name__)

# constraint comes from constant
Response = namedtuple('Response', 'src_name, src_owner, dest_name, tau_days, scale, start')

State = namedtuple('State', 'prev_time, prev_value')


class ImpulseCalculator(CompositeCalculatorMixin, UniProcCalculator):

    def __init__(self, *args, responses=None, impulse=None, **kargs):
        self.responses_ref = self._assert('responses', responses)
        self.impulse_ref = self._assert('impulse', impulse)
        self.__state = {}
        self.__impulse_id = None
        self.__prev_source_id = None
        super().__init__(*args, **kargs)

    def _startup(self, s):
        super()._startup(s)
        self.constants = [Constant.get(s, response) for response in self.responses_ref]
        self.responses = [Response(**loads(constant.at(s).value)) for constant in self.constants]
        self.impulse = HRImpulse(**loads(Constant.get(s, self.impulse_ref).at(s).value))

    def _missing(self, s):
        self.__impulse_id = s.query(StatisticName.id). \
            filter(StatisticName.name == self.impulse.dest_name,
                   StatisticName.owner == self.owner_in).scalar()
        log.debug(f'Impulse ID: {self.__impulse_id}')
        try:
            dates = super()._missing(s)
            if dates:
                # need to delete forwards just in case there's some weird gap
                log.info('Need to delete forwards for Impulse')
                self._delete_time_range(s, local_date_to_time(dates[0]), 0)
                dates = super()._missing(s)
                start, finish = dates[0], dates[-1]
                # may need to bridge backwards
                prev = self._previous_source(s)
                if prev:
                    start = time_to_local_date(prev.start) + dt.timedelta(days=1)
                return list(self.__days(start, finish))
            else:
                return []
        finally:
            self.__set_state(s)

    def __days(self, start, finish):
        day = dt.timedelta(days=1)
        while start <= finish:
            yield start
            start += day

    def _previous_source(self, s):
        return s.query(ActivityJournal). \
            join(CompositeComponent, CompositeComponent.input_source_id == ActivityJournal.id). \
            join(StatisticJournal, CompositeComponent.output_source_id == StatisticJournal.source_id). \
            join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
            filter(StatisticName.owner == self.owner_out). \
            order_by(desc(ActivityJournal.start)).limit(1).one_or_none()

    def _unused_sources_given_used(self, s, used_sources):
        sources = s.query(ActivityJournal). \
            join(StatisticJournal, StatisticName). \
            filter(~ActivityJournal.id.in_(used_sources),
                   StatisticName.id == self.__impulse_id)
        start, finish = self._start_finish(local_date_to_time)
        if start:
            sources = sources.filter(ActivityJournal.finish >= start)
        if finish:
            sources = sources.filter(ActivityJournal.start <= finish)
        log.debug(f'Unused query: {sources}')
        return sources.all()

    def __set_state(self, s):
        for response, constant in zip(self.responses, self.constants):
            start, finish = self._start_finish(local_date_to_time)
            name = response.dest_name
            activity_group = constant.statistic_name.constraint
            prev = s.query(StatisticJournal). \
                join(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.owner == self.owner_out,
                       StatisticName.constraint == activity_group). \
                order_by(desc(StatisticJournal.time)).limit(1).one_or_none()
            prev_time = prev.time if prev else start
            prev_value = prev.value if prev else 1e-20  # avoid zero as it gives numerical issues later
            self.__state[name] = State(prev_time=prev_time, prev_value=prev_value)
            log.debug(f'State for {name}: {self.__state[name]}')
            if not self.__prev_source_id:
                self.__prev_source_id = prev.source_id if prev else None
                log.debug(f'Source ID: {self.__prev_source_id}')

    def _read_data(self, s, start, finish):
        impulses, t, source_ids = {}, _tables(), set()
        for response, constant in zip(self.responses, self.constants):
            name = response.dest_name
            stmt = select([t.sj.c.time, t.sj.c.source_id, t.sjf.c.value]). \
                select_from(t.sj.join(t.sjf)). \
                where(and_(t.sj.c.statistic_name_id == self.__impulse_id,
                           t.sj.c.time >= start,
                           t.sj.c.time < finish)). \
                order_by(t.sj.c.time)
            rows = list(s.connection().execute(stmt))
            impulses[name] = [(row[0], row[2]) for row in rows]
            source_ids.update(row[1] for row in rows)
        if self.__prev_source_id:
            source_ids.add(self.__prev_source_id)
        return source_ids, impulses

    def _calculate_results(self, s, source, data, loader, start, finish):
        for response, constant in zip(self.responses, self.constants):
            name = response.dest_name
            impulses = data[name]
            prev_time, value = self.__state[name]
            for time, impulse in self.__pad(impulses, start, finish):
                if prev_time:
                    dt = (time - prev_time).total_seconds()
                    value *= exp(-dt / (response.tau_days * 24 * 60 * 60))
                value += response.scale * impulse
                # log.debug(f'{time}: {value}')
                loader.add(response.dest_name, None, MAX, constant.statistic_name.constraint, source, value, time,
                           StatisticJournalFloat)
                prev_time = time
            self.__state[name] = self.__state[name]._replace(prev_time=prev_time, prev_value=value)
        self.__prev_source_id = source.id
        log.debug(f'Source ID: {self.__prev_source_id}')

    def __pad(self, impulses, start, finish):
        delta = dt.timedelta(hours=1)  # every hour is assumed by data.frame.std_health_statistics
        for (time, impulse) in impulses:
            while start < time:
                yield start, 0
                start += delta
            yield time, impulse
            start = time + delta
        while start < finish:
            yield start, 0
            start += delta




from collections import namedtuple
from logging import getLogger
from math import pi

import numpy as np
import pandas as pd
import scipy as sp

from .fit import fit
from ..stoats.names import *
from ..stoats.names import _d, _sqr, _avg

log = getLogger(__name__)
RAD_TO_DEG = 180 / pi


def median_dt(stats):
    return pd.Series(stats.index).diff().median().total_seconds()


def linear_resample(df, start=None, finish=None, dt=None, with_timestamp=None, keep_nan=True):
    if with_timestamp is None: with_timestamp = TIMESPAN_ID in df.columns
    dt = dt or median_dt(df)
    start = start or df.index.min()
    finish = finish or df.index.max()
    even = pd.DataFrame({'keep': True}, index=pd.date_range(start=start, end=finish, freq=f'{dt}S'))
    both = df.join(even, how='outer', sort=True)
    both.loc[both['keep'] != True, ['keep']] = False  # not sure this is needed, but avoid interpolating to true
    both.interpolate(method='index', limit_area='inside', inplace=True)
    resampled = both.loc[both['keep'] == True].drop(columns=['keep'])
    resampled[TIME] = resampled.index
    resampled[DELTA_TIME] = resampled[TIME].diff()
    if with_timestamp:
        if keep_nan:
            resampled.loc[~resampled[TIMESPAN_ID].isin(df[TIMESPAN_ID].unique())] = np.nan
        else:
            resampled = resampled.loc[resampled[TIMESPAN_ID].isin(df[TIMESPAN_ID].unique())]
    return resampled


def add_differentials(df):
    return _add_differentials(df, SPEED, DISTANCE, ELEVATION, SPEED, SPEED_2, LATITUDE, LONGITUDE)


def add_air_speed(df, wind_speed=0, wind_heading=0):
    df[AIR_SPEED] = df[SPEED] + wind_speed * np.cos((df[HEADING] - wind_heading) / RAD_TO_DEG)
    return _add_differentials(df, AIR_SPEED)


def _add_differentials(df, speed, *names):

    speed_2 = _sqr(speed)
    df[speed_2] = df[speed] ** 2

    def diff():
        for _, old_span in df.groupby(TIMESPAN_ID):

            # discard leading and trailing na
            subset = old_span[list(names)].isna().any(axis=1).replace(True, np.nan)
            start, finish = subset.first_valid_index(), subset.last_valid_index()
            if start and finish:
                old_span = old_span.loc[start:finish]

                if all(len(old_span[name]) == len(old_span[name].dropna()) for name in names):
                    new_span = pd.DataFrame(index=old_span.index)
                    for col in names:
                        new_span[_d(col)] = old_span[col].diff()
                    if HEADING not in old_span.columns:
                        new_span[HEADING] = np.arctan2(new_span[_d(LONGITUDE)], new_span[_d(LATITUDE)]) * RAD_TO_DEG
                    avg_speed_2 = [(a**2 + a*b + b**2)/3 for a, b in zip(old_span[speed], old_span[speed][1:])]
                    new_span[_avg(speed_2)] = [np.nan] + avg_speed_2
                    yield new_span

    spans = list(diff())
    if len(spans):
        extra = pd.concat(spans)
        return df.drop(columns=list(extra.columns), errors='ignore').join(extra)
    else:
        raise PowerException('Missing data - found no spans without NANs')


def add_energy_budget(df, m, g=9.8):
    # if DELTA_ELEVATION is +ve we've gone uphill.  so this is the total amount of energy
    # gained in this segment.
    df[DELTA_ENERGY] = m * (df[DELTA_SPEED_2] / 2 + df[DELTA_ELEVATION] * g)
    return df


def add_cda_estimate(df, p=1.225):
    # https://www.cyclingpowerlab.com/CyclingAerodynamics.aspx
    # assume that all energy lost (-ve gain) is due to air resistance.
    df[CDA] = -df[DELTA_ENERGY] / (p * df[AVG_AIR_SPEED_2] * df[DELTA_DISTANCE] * 0.5)
    return df


def add_crr_estimate(df):
    # assume that all energy lost is due to rolling resistance
    df[CRR] = -df[DELTA_ENERGY] / df[DELTA_DISTANCE]
    return df


def add_loss_estimate(df, cda=0.45, crr=0, p=1.225):
    # this is the energy spent on air and rolling resistance
    df[LOSS] = (cda * p * df[AVG_AIR_SPEED_2] * 0.5 + crr) * df[DELTA_DISTANCE]
    return df


def add_power_estimate(df):
    # power input must balance the energy budget.
    df[POWER] = (df[DELTA_ENERGY] + df[LOSS]) / df[DELTA_TIME].dt.total_seconds()
    df[POWER].clip(lower=0, inplace=True)
    if CADENCE in df.columns: df.loc[df[CADENCE] < 1, [POWER]] = 0
    df.loc[df[POWER].isna(), [POWER]] = 0
    energy = (df[POWER].iloc[1:] * df[DELTA_TIME].iloc[1:]).cumsum()
    df[ENERGY] = 0
    df.loc[1:, [ENERGY]] = energy
    return df


def add_modeled_hr(df, window, slope, delay):
    window = int(0.5 + window / median_dt(df))
    df[DETRENDED_HEART_RATE] = df[HEART_RATE] - df[HEART_RATE].rolling(window, center=True, min_periods=1).median()
    predicted = (df[POWER] * slope).ewm(halflife=delay).mean()
    df[PREDICTED_HEART_RATE] = predicted - predicted.rolling(window, center=True, min_periods=1).median()
    return df


def measure_initial_delay(df, dt=None, col1=HEART_RATE, col2=POWER, n=20):
    dt = dt or median_dt(df)
    correln = [(i, df[col1].corr(df[col2].shift(freq=f'{i * dt}S'))) for i in range(-n, n + 1)]
    correln = sorted(correln, key=lambda c: c[1], reverse=True)
    return dt * correln[0][0]


def measure_initial_scaling(df):
    delay = measure_initial_delay(df)
    if delay < 0: raise PowerException('Cannot estimate delay (insufficient data?)')
    df[DELAYED_POWER] = df[POWER].shift(freq=f'{delay}S')
    clean = df.loc[:, (DELAYED_POWER, HEART_RATE)].dropna()
    fit = sp.stats.linregress(x=clean[DELAYED_POWER], y=clean[HEART_RATE])
    log.debug(f'Initial fit {fit}')
    return fit.slope, fit.intercept,  delay


class PowerException(Exception): pass


# the 13.5 delay is from fitting my rides til 2019
# it seems oddly low to me - i wonder if i have an error confusing bins and time
PowerModel = namedtuple('PowerModel', 'cda, crr, slope, window, delay, m,  wind_speed, wind_heading',
                             defaults=[0.5, 0,   200,   60*60,  13.5,  70, 0,          0])


def evaluate(df, model, quiet=True):
    if not quiet: log.debug(f'Evaluating {model}')
    df = add_energy_budget(df, model.m)
    df = add_air_speed(df, model.wind_speed, model.wind_heading)
    df = add_loss_estimate(df, model.cda, model.crr)
    df = add_power_estimate(df)
    return df


MIN_DELAY = 1


def fit_power(df, model, *vary):

    log.debug(f'Fit power: varying {vary}')
    df = evaluate(df, model)
    dt = median_dt(df)
    slope, intercept, delay = measure_initial_scaling(df)
    # we ignore intercept because removing the median makes it very weakly constrained
    model = model._replace(slope=slope, delay=delay)
    log.debug(f'Fit power: initial model {model}')

    # the internal delay is continuous
    # model delay is MIN_DELAY when the internal delay is zero, and otherwise increases

    def forwards(kargs):
        if 'delay' in kargs:
            kargs['delay'] = kargs['delay'] - MIN_DELAY
        return kargs

    def backwards(kargs):
        if 'delay' in kargs:
            kargs['delay'] = abs(kargs['delay']) + MIN_DELAY
        return kargs

    def evaluate_and_extend(df, model):
        df = evaluate(df, model)
        df = add_modeled_hr(df, model.window, model.slope, model.delay)
        return df

    model = fit(DETRENDED_HEART_RATE, PREDICTED_HEART_RATE, df, model, evaluate_and_extend,
                *vary, forwards=forwards, backwards=backwards)

    log.debug(f'Fit power: model before fixing {model}')
    if model.wind_speed < 0:
        model = model._replace(wind_speed=abs(model.wind_speed), wind_heading=model.wind_heading+180)
    model = model._replace(wind_heading=model.wind_heading % 360)

    log.debug(f'Fit power: final model {model}')
    return model
